part of insync;

abstract class AbstractCollection<V> extends Owner {
  int get length;
  bool get isEmpty;
  V get first;

  bool containsValue(V value);
  void removeValue(V value);
  void removeValues(List<V> values);
  void tryClear();
  void clear();
  AbstractCollection<V> $clear();

  bool every(bool f(V value));
  dynamic filter(bool f(V value));
  AbstractCollection<V> $filter(bool f(V value));
  dynamic map(dynamic f(V value));
  AbstractCollection<dynamic> $map(dynamic f(V value));

  AbstractCollection<dynamic> createEmpty();
  AbstractList<dynamic> createEmptyList();
  AbstractMap<dynamic, dynamic> createEmptyMap();
  AbstractSet<dynamic> createEmptySet();

  AbstractCollectionMapper<V, dynamic, AbstractCollection<V>, AbstractCollection<dynamic>> createMapper(
      {dynamic create(V data), void destroy(dynamic item, V data)});
  AbstractCollectionFilterer<V, AbstractCollection<V>> createFilterer(
      {bool filter(V data)});
  AbstractCollectionObserver<V, AbstractCollection<V>> createObserver(
      {void add(V value), void remove(V value), void clear(List<V> values)});
  AbstractCollectionOrderer<V, AbstractCollection<V>> createOrderer(
      {AbstractList<V> target});
  AbstractCollectionSorterComparing<V, AbstractCollection<V>> createSorterComparing(
      {AbstractList<V> target, int compare(V x, V y)});
  AbstractCollectionIndexer<dynamic, V, AbstractCollection<V>> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)});
  AbstractCollectionHeaper<V, AbstractCollection<V>> createHeaper(
      {AbstractSet<V> target});

  bool _ownsValues = false;
  void ownValues() {
    _ownsValues = true;
  }

  void destroy() {
    tryClear();
    super.destroy();
  }

  bool some(bool f(V value)) {
    return !every((V v) => (f(v) == false));
  }

  void each(void f(V value)) {
    every((V v) { f(v); return true; });
  }

  V search(bool f(V value)) {
    V result;
    every((V v) { if (f(v) != false) { result = v; return false; } return true; });
    return result;
  }

  List<V> toSorted([Comparable callback(V value), int order = 1]) {
    if (callback == null) {
      callback = (V v) => v;
    }
    List<Pair<V, Comparable>> pairs = new List();
    every((V v) { pairs.add(new Pair(v, callback(v))); });
    pairs.sort((Pair<V, Comparable> x, Pair<V, Comparable> y) => order * x.y.compareTo(y.y));
    return pairs.map((Pair<V, Comparable> t) => t.x).toList(growable: true);
  }

  SimpleList<V> $toSorted([Comparable callback(V value), int order = 1]) {
    return new SimpleList.adapter(toSorted(callback, order));
  }

  List<V> toSortedComparing(int compare(V x, V y), [int order = 1]) {
    List<V> values = toList();
    values.sort((V x, V y) => order * compare(x, y));
    return values;
  }

  SimpleList<V> $toSortedComparing(int compare(V x, V y), [int order = 1]) {
    return new SimpleList.adapter(toSortedComparing(compare, order));
  }

  Map<dynamic, V> index(dynamic f(V value)) {
    Map<dynamic, V> result = new HashMap();
    every((V v) { dynamic key = f(v); if (key != null) { result[key] = v; } return true; });
    return result;
  }

  SimpleMap<dynamic, V> $index(dynamic f(V value)) {
    return new SimpleMap.adapter(index(f));
  }

  List<V> toList() {
    List<V> result = new List<V>();
    every((V v) { result.add(v); return true; });
    return result;
  }

  SimpleList<V> $toList() {
    return new SimpleList.adapter(toList());
  }

  Set<V> toSet() {
    Set<V> result = new Set<V>();
    every((V v) { result.add(v); return true; });
    return result;
  }

  SimpleSet<V> $toSet() {
    return new SimpleSet.adapter(toSet());
  }

  List<V> asList() {
    return toList();
  }

  AbstractList<V> $asList() {
    return $toList();
  }

  Set<V> asSet() {
    return toSet();
  }

  AbstractSet<V> $asSet() {
    return $toSet();
  }
}