part of insync;

abstract class AbstractList<V> extends IndexedCollection<int, V> {
  List<V> _list = new List();
  List<V> _constantList;

  AbstractList<dynamic> createEmpty();

  AbstractList() {
    _constantList = new ConstantList(_list);
  }

  AbstractList.from(List<V> values) {
    _constantList = new ConstantList(_list);
    _list.addAll(values);
  }

  AbstractList.adapter(this._list) {
    _constantList = new ConstantList(_list);
  }

  AbstractList.filled(int length, V value) : _list = new List.filled(length, value) {
    _constantList = new ConstantList(_list);
  }

  List<V> get list => _constantList;
  int get firstKey => isEmpty ? null : 0;
  int get lastKey => _list.length - 1;
  V get last => _list[lastKey];
  int get length => _list.length;
  bool get isEmpty => _list.isEmpty;
  V operator [](int index) => _list[index];

  List<int> getKeys() {
    return new List<int>.generate(_list.length, (int i) => i, growable: true);
  }

  bool every2(bool f(V value, int index)) {
    for (int i = 0; i < _list.length; ++i) {
      if (f(_list[i], i) == false) {
        return false;
      }
    }
    return true;
  }

  List<V> filter(bool f(V value)) {
    return filter2((V v, int i) => f(v));
  }

  SimpleList<V> $filter(bool f(V value)) {
    return new SimpleList.adapter(filter(f));
  }

  List<V> filter2(bool f(V value, int index)) {
    List<V> result = new List();
    every2((V v, int i) { if (f(v, i) != false) { result.add(v); }});
    return result;
  }

  SimpleList<V> $filter2(bool f(V value, int index)) {
    return new SimpleList.adapter(filter2(f));
  }

  List<dynamic> map(dynamic f(V value)) {
    return map2((V v, int i) => f(v));
  }

  SimpleList<dynamic> $map(dynamic f(V value)) {
    return new SimpleList.adapter(map(f));
  }

  List<dynamic> map2(dynamic f(V value, int index)) {
    List<dynamic> result = new List();
    every2((V v, int i) { result.add(f(v, i)); });
    return result;
  }

  SimpleList<dynamic> $map2(dynamic f(V value, int index)) {
    return new SimpleList.adapter(map2(f));
  }

  List<V> toList() {
    return new ListUtil<V>().clone(_list);
  }

  Set<V> toSet() {
    return list.toSet();
  }

  List<V> asList() {
    return list;
  }

  AbstractList<V> $asList() {
    return this;
  }

  void add(V value, [int index]) {
    tryAdd(value, index);
  }

  bool tryAdd(V value, [int index]) {
    return tryAddAll([value], index);
  }

  void addAll(List<V> values, [int index]) {
    tryAddAll(values, index);
  }

  bool tryAddAll(List<V> values, [int index]) {
    if (index == null) {
      index = length;
    }
    return (trySplice([], [new IndexValues(index, values)]) != null) ? true : null;
  }

  Wrap<V> trySet(V value, int index) {
    Wrap<V> oldWrap = new ListUtil<V>().trySet(_list, value, index);
    if ((oldWrap != null) && _ownsValues) {
      (oldWrap.value as Destroyable).destroy();
    }
    return oldWrap;
  }

  V tryRemove(int index) {
    List<V> result = tryRemoveAll(index, 1);
    return (result != null) ? result[0] : null;
  }

  List<V> removeAll(int index, int count) {
    List<V> result = tryRemoveAll(index, count);
    return (result != null) ? result : new List<V>();
  }

  SimpleList<V> $removeAll(int index, int count) {
    return new SimpleList.adapter(removeAll(index, count));
  }

  List<V> tryRemoveAll(int index, int count) {
    ListSpliceResult<V> result = trySplice([new IndexCount(index, count)], []);
    return (result != null) ? result.removedValuesList[0].values : null;
  }

  void removeValues(List<V> values) {
    Set<V> valueSet = values.toSet();
    List<V> newValues = filter((V value) => !valueSet.contains(value));
    performFilter(newValues);
  }

  V move(int fromIndex, int toIndex) {
    tryMove(fromIndex, toIndex);
    return this[toIndex];
  }

  V tryMove(int fromIndex, int toIndex) {
    return new ListUtil<V>().tryMove(_list, fromIndex, toIndex);
  }

  List<V> clear() {
    List<V> result = tryClear();
    return (result != null) ? result : new List<V>();
  }

  SimpleList<V> $clear() {
    return new SimpleList.adapter(clear());
  }

  List<V> tryClear() {
    List<V> values = new ListUtil<V>().tryClear(_list);
    if ((values != null) && _ownsValues) {
      values.forEach(Destroyable.destroyObject);
    }
    return values;
  }

  ListSpliceResult<V> splice(List<IndexCount> removeParamsList, List<IndexValues<V>> addParamsList) {
    ListSpliceResult<V> result = trySplice(removeParamsList, addParamsList);
    return (result != null) ? result : new ListSpliceResult(new ListUtil<V>().clone(_list));
  }

  ListSpliceResult<V> trySplice(List<IndexCount> removeParamsList,
      List<IndexValues<V>> addParamsList) {
    ListSpliceResult<V> spliceResult = new ListUtil<V>().trySplice(
        _list, removeParamsList, addParamsList);
    if ((spliceResult != null) && _ownsValues) {
      spliceResult.removedValues.forEach(Destroyable.destroyObject);
    }
    return spliceResult;
  }

  void reorder(List<int> indexList) {
    tryReorder(indexList);
  }

  List<V> tryReorder(List<int> indexList) {
    return new ListUtil<V>().tryReorder(_list, indexList);
  }

  V pop() {
    return isEmpty ? null : tryRemove(length - 1);
  }

  ListSpliceParams<V> detectSplice(List<V> newValues) {
    return new ListUtil<V>().detectSplice(_list, newValues);
  }

  List<IndexCount> detectFilter(List<V> newValues) {
    return new ListUtil<V>().detectFilter(_list, newValues);
  }

  List<int> detectReorder(List<V> newValues) {
    return new ListUtil<V>().detectReorder(_list, newValues);
  }

  List<int> detectSort([Comparable callback(V value, int index), int order = 1]) {
    return ListUtil.invert(getSortingKeys(callback, order));
  }

  List<int> detectSortComparing(int compare(V v1, int i1, V v2, int i2), [int order = 1]) {
    return ListUtil.invert(getSortingKeysComparing(compare, order));
  }

  void performSplice(List<V> newValues) {
    ListSpliceParams<V> params = detectSplice(newValues);
    if (params != null) {
      trySplice(params.removeParamsList, params.addParamsList);
    }
  }

  void performFilter(List<V> newValues) {
    List<IndexCount> params = detectFilter(newValues);
    if (params != null) {
      trySplice(params, []);
    }
  }

  void performReorder(List<V> newValues) {
    List<int> indexList = detectReorder(newValues);
    if (indexList != null) {
      tryReorder(indexList);
    }
  }

  void sort([Comparable callback(V value), int order = 1]) {
    sort2((V v, int i) => callback(v), order);
  }

  void sort2([Comparable callback(V value, int index), int order = 1]) {
    List<int> indexList = detectSort(callback, order);
    if (indexList != null) {
      tryReorder(indexList);
    }
  }

  void sortComparing(int compare(V x, V y), [int order = 1]) {
    sortComparing2((V v1, int i1, V v2, int i2) => compare(v1, v2), order);
  }

  void sortComparing2(int compare(V v1, int i1, V v2, int i2), [int order = 1]) {
    List<int> indexList = detectSortComparing(compare, order);
    if (indexList != null) {
      tryReorder(indexList);
    }
  }

  AbstractListMapper<V, dynamic> createMapper({dynamic create(V data),
    void destroy(dynamic item, V data)}) {
    return new AbstractListMapper<V, dynamic>(this, create: create, destroy: destroy);
  }

  AbstractListMapper<V, dynamic> createMapperTo({AbstractList<dynamic> target,
    dynamic create(V data), void destroy(dynamic item, V data)}) {
    return new AbstractListMapper<V, dynamic>(this, target: target, create: create, destroy: destroy);
  }

  AbstractListFilterer<V> createFilterer({bool filter(V value)}) {
    return new AbstractListFilterer<V>(this, filter: filter);
  }

  AbstractListFilterer<V> createFiltererTo({AbstractList<V> target, bool filter(V value)}) {
    return new AbstractListFilterer<V>(this, target: target, filter: filter);
  }

  AbstractListObserver<V> createObserver({void add(V value), void remove(V value),
      void clear(List<V> values), void change()}) {
    return new AbstractListObserver<V>(this,
        add: add, remove: remove, clear: clear, change: change);
  }

  AbstractListOrderer<V> createOrderer({AbstractList<V> target}) {
    return new AbstractListOrderer<V>(this, target: target);
  }

  AbstractListSorterComparing<V> createSorterComparing({AbstractList<V> target,
      int compare(V x, V y)}) {
    return new AbstractListSorterComparing<V>(this, target: target, compare: compare);
  }

  AbstractListIndexer<dynamic, V> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)}) {
    return new AbstractListIndexer<dynamic, V>(this, target: target, getKey: getKey);
  }

  AbstractListHeaper<V> createHeaper({AbstractSet<V> target}) {
    return new AbstractListHeaper<V>(this, target: target);
  }

  AbstractListInserter<V> createInserter({void add(V value, int index),
      void remove(V value, int index), void clear(List<V> values)}) {
    return new AbstractListInserter<V>(this, add: add, remove: remove, clear: clear);
  }

  AbstractListMerger<dynamic> createMerger({AbstractList<dynamic> target}) {
    return new AbstractListMerger<dynamic>(this, target: target);
  }

  _AbstractListMergerBunch<dynamic> _createMergerBunch(AbstractListMerger<dynamic> merger) {
    return new _AbstractListMergerBunch<dynamic>();
  }

  AbstractListReverser<V> createReverser({AbstractList<V> target}) {
    return new AbstractListReverser<V>(this, target: target);
  }

  bool equals(List<V> other) {
    return new ListUtil<V>().equal(_list, other);
  }

  int indexOf(V value, [int start]) {
    return _list.indexOf(value, start);
  }

  int binarySearch(bool isBigEnough(V value)) {
    return new ListUtil<V>().binarySearch(_list, isBigEnough);
  }
}