part of insync;

class AbstractListFilterer<V> extends AbstractCollectionFilterer<V, AbstractList<V>> {
  List<int> _filtered = new List<int>();

  AbstractListFilterer(AbstractList<V> source, {AbstractList<V> target, bool filter(V value)}) :
    super(source, target: target, filter: filter) {
    _splice([], [new IndexValues<V>(0, _source.list)]);
  }

  void destroy() {
    _target.tryClear();
    super.destroy();
  }

  int _countFiltered(int index, int count) {
    int result = 0;
    for (int i = 0; i < count; ++i) {
      result += _filtered[index + i];
    }
    return result;
  }

  void _splice(List<IndexValues<V>> removedValuesList, List<IndexValues<V>> addedValuesList) {
    int sourceIndex = 0;
    int targetIndex = 0;
    List<IndexCount> removeParamsList = removedValuesList.map((IndexValues<V> indexValues) {
      targetIndex += _countFiltered(sourceIndex, indexValues.index - sourceIndex);
      int count = _countFiltered(indexValues.index, indexValues.values.length);
      IndexCount params = new IndexCount(targetIndex, count);
      sourceIndex = indexValues.index + indexValues.values.length;
      targetIndex += count;
      return params;
    }).toList(growable: false);
    new SimpleList<V>.adapter(_filtered).trySplice(
        removedValuesList.map((IndexValues<V> v) => v.toIndexCount()).toList(growable: false), []);

    sourceIndex = 0;
    targetIndex = 0;
    List<IndexValues<V>> addParamsList = addedValuesList.map((IndexValues<V> indexValues) {
      targetIndex += _countFiltered(sourceIndex, indexValues.index - sourceIndex);
      List<V> values = [];
      List<int> filtered = indexValues.values.map((V value) {
        if (_filterValue(value) == false) {
          return 0;
        }
        values.add(value);
        return 1;
      }).toList(growable: false);
      IndexValues<V> params = new IndexValues<V>(targetIndex, values);
      _filtered.insertAll(indexValues.index, filtered);
      sourceIndex = indexValues.index + filtered.length;
      targetIndex += values.length;
      return params;
    }).toList(growable: false);

    target.trySplice(removeParamsList, addParamsList);
  }

  void reconfigure({bool filter(V value)}) {
    if (filter == null) {
      return;
    }
    _filterValue = filter;
    refilter();
  }

  void refilterAt(int sourceIndex) {
    V value = _source[sourceIndex];
    assert(value != null);
    bool good = _filterValue(value) != false;
    int targetIndex = _countFiltered(0, sourceIndex);
    if (_filtered[sourceIndex] == 0) {
      if (good) {
        _filtered[sourceIndex] = 1;
        _target.add(value, targetIndex);
      }
    } else {
      if (!good) {
        _filtered[sourceIndex] = 0;
        _target.remove(targetIndex);
      }
    }
  }

  void refilterValue(V value) {
    int index = _source.indexOf(value);
    if (index != -1) {
      refilterAt(index);
    }
  }

  void refilter() {
    List<int> newFiltered = _source.map((V value) => (_filterValue(value) != false) ? 1 : 0);
    IndexCount removeParams;
    List<IndexCount> removeParamsList = [];

    void flushRemove() {
      if (removeParams != null) {
        removeParamsList.add(removeParams);
        removeParams = null;
      }
    }

    int targetIndex = 0;
    _source.every2((V value, int index) {
      if (_filtered[index] == 0) {
        return;
      }
      if (newFiltered[index] == 0) {
        if (removeParams == null) {
          removeParams = new IndexCount(targetIndex, 0);
        }
        ++removeParams.count;
        _filtered[index] = 0;
      } else {
        flushRemove();
      }
      ++targetIndex;
    });

    flushRemove();

    IndexValues<V> addParams;
    List<IndexValues<V>> addParamsList = [];

    void flushAdd() {
      if (addParams != null) {
        addParamsList.add(addParams);
        addParams = null;
      }
    }

    targetIndex = 0;
    _source.every2((V value, int index) {
      if (_filtered[index] == 1) {
        flushAdd();
        ++targetIndex;
        return;
      }
      if (newFiltered[index] == 1) {
        if (addParams == null) {
          addParams = new IndexValues<V>(targetIndex, []);
        }
        addParams.values.add(value);
        _filtered[index] = 1;
        ++targetIndex;
      } else {
        flushAdd();
      }
    });

    flushAdd();

    _filtered = newFiltered;
    _target.trySplice(removeParamsList, addParamsList);
  }
}