part of insync;

class AbstractListMerger<V> extends Owner {
  AbstractList<AbstractList<V>> _source;
  AbstractList<AbstractList<V>> get source => _source;

  AbstractList<V> _target;
  AbstractList<V> get target => _target;

  AbstractListMerger(this._source, {AbstractList<V> target}) {
    assert(_source != null);
    _target = (target != null) ? target : _createTarget()..ownBy(this);
    own(_source.createMapper(
      create: (AbstractList<V> values) => values._createMergerBunch(this),
      destroy: Destroyable.destroyFromMapper
    ));
    _target.tryAddAll(_getAllValues());
  }

  void destroy() {
    _target.tryClear();
    super.destroy();
  }

  AbstractList<V> _createTarget() {
    return _source.some((AbstractList<V> values) => values is ObservableList) ?
        new ObservableList() : new SimpleList();
  }

  List<V> _getAllValues() {
    return _merge(_source.list);
  }

  List<V> _merge(List<AbstractList<V>> lists) {
    List<V> values = new List<V>.filled(_count(lists), null);
    int iValues = 0;
    for (int i = 0, l = lists.length; i < l; ++i) {
      List<V> list = lists[i].list;
      for (int j = 0, m = list.length; j < m; ++j) {
        values[iValues++] = list[j];
      }
    }
    return values;
  }

  int _count(List<AbstractList<V>> lists, [int index, int length]) {
    if (index == null) {
      index = 0;
    }
    if (length == null) {
      length = lists.length - index;
    }
    int count = 0;
    for (int i = 0; i < length; ++i) {
      count += lists[index + i].length;
    }
    return count;
  }
}