part of insync;

class IndexCount {
  int index;
  int count;

  IndexCount(this.index, this.count) {}
}

class IndexValues<V> {
  int index;
  List<V> values;

  SimpleList<V> get $values => new SimpleList.adapter(values);

  IndexValues(this.index, this.values) {}

  IndexCount toIndexCount() {
    return new IndexCount(index, values.length);
  }
}

class ListSpliceParams<V> {
  List<IndexCount> removeParamsList;
  List<IndexValues<V>> addParamsList;

  ListSpliceParams() : removeParamsList = new List(), addParamsList = new List() {}

  bool get isEmpty {
    return removeParamsList.isEmpty && addParamsList.isEmpty;
  }
}

class ListSpliceResult<V> {
  List<V> oldValues;
  SimpleList<V> get $oldValues => new SimpleList.adapter(oldValues);

  List<IndexValues<V>> removedValuesList;
  List<IndexValues<V>> addedValuesList;

  List<V> _removedValues;
  List<V> _addedValues;
  List<IndexCount> _removeParamsList;

  ListSpliceResult(this.oldValues) :
      removedValuesList = new List(),
      addedValuesList = new List() {}

  List<V> get removedValues {
    if (_removedValues == null) {
      _removedValues = new ListUtil<V>().merge(
          removedValuesList.map((IndexValues<V> list) => list.values).toList(growable: true));
    }
    return _removedValues;
  }

  SimpleList<V> get $removedValues => new SimpleList.adapter(removedValues);

  List<V> get addedValues {
    if (_addedValues == null) {
      _addedValues = new ListUtil<V>().merge(
          addedValuesList.map((IndexValues<V> list) => list.values).toList(growable: true));
    }
    return _addedValues;
  }

  SimpleList<V> get $addedValues => new SimpleList.adapter(addedValues);

  List<IndexCount> get removeParamsList {
    if (_removeParamsList == null) {
      _removeParamsList = removedValuesList.map(
          (IndexValues list) => list.toIndexCount()).toList(growable: true);
    }
    return _removeParamsList;
  }

  bool get isEmpty {
    return removedValuesList.isEmpty && addedValuesList.isEmpty;
  }
}

class ListUtil<V> {
  List<V> clone(List<V> list) {
    return new List()..addAll(list);
  }

  bool equal(List<V> x, List<V> y) {
    if (identical(x, y)) {
      return true;
    }
    if (x.length != y.length){
      return false;
    }
    for (int i = 0, l = x.length; i < l; ++i) {
      if (x[i] != y[i]) {
        return false;
      }
    }
    return true;
  }

  bool tryAddAll(List<V> target, List<V> values, [int index]) {
    if (values.isEmpty) {
      return null;
    }
    if (index == null) {
      index = target.length;
    }
    target.insertAll(index, values);
    return true;
  }

  List<V> tryRemoveAll(List<V> target, int index, int count) {
    assert((index >= 0) && (count >= 0) && (index + count < target.length));
    if (count == 0) {
      return null;
    }
    List<V> values = target.getRange(index, index + count).toList(growable: false);
    target.removeRange(index, index + count);
    return values;
  }

  V trySet(List<V> target, V value, int index) {
    assert(value != null);
    assert(index >= 0);
    assert(index < target.length);
    V oldValue = target[index];
    if (value == oldValue) {
      return null;
    }
    target[index] = value;
    return oldValue;
  }

  V tryMove(List<V> target, int fromIndex, int toIndex) {
    assert((fromIndex >= 0) && (fromIndex < target.length));
    assert((toIndex >= 0) && (toIndex < target.length));
    if (fromIndex == toIndex) {
      return null;
    }
    V value = target[fromIndex];
    target.removeAt(fromIndex);
    target.insert(toIndex, value);
    return value;
  }

  List<V> tryClear(List<V> target) {
    if (target.isEmpty) {
      return null;
    }
    List<V> values = target.getRange(0, target.length).toList(growable: true);
    target.clear();
    return values;
  }

  /**
   * @todo Assert out of bounds stuff
   */
  ListSpliceResult<V> trySplice(List<V> target, List<IndexCount> removeParamsList,
      List<IndexValues<V>> addParamsList) {
    ListSpliceResult<V> spliceResult = new ListSpliceResult(clone(target));
    for (int i = removeParamsList.length - 1; i >= 0; --i) {
      IndexCount params = removeParamsList[i];
      int index = params.index;
      List<V> values = tryRemoveAll(target, index, params.count);
      if (values == null) {
        continue;
      }
      spliceResult.removedValuesList.add(new IndexValues(index, values));
    }
    for (int i = 0; i < addParamsList.length; ++i) {
      IndexValues<V> params = addParamsList[i];
      if (tryAddAll(target, params.values, params.index) == null) {
        continue;
      }
      spliceResult.addedValuesList.add(params);
    }
    if (spliceResult.isEmpty) {
      return null;
    }
    spliceResult.removedValuesList = spliceResult.removedValuesList.reversed.toList(growable: true);
    return spliceResult;
  }

  List<V> tryReorder(List<V> target, List<int> indexList) {
    assert(target.length == indexList.length);
    List<int> indexListSorted = new ListUtil<int>().clone(indexList);
    indexListSorted.sort();
    assert(isIdentity(indexListSorted));
    if (isIdentity(indexList)) {
      return null;
    }
    List<V> oldValues = clone(target);
    for (int i = 0; i < target.length; ++i) {
      target[indexList[i]] = oldValues[i];
    }
    return oldValues;
  }

  ListSpliceParams<V> detectSplice(List<V> oldValues, List<V> newValues) {
    ListSpliceParams<V> spliceParams = new ListSpliceParams();
    Map<V, int> oldIndexMap = getIndexMap(oldValues);
    int nextOldIndex = 0;
    int offset = 0;
    List<V> newValueBuffer = new List();

    void flush() {
      if (newValueBuffer.isEmpty) {
        return;
      }
      spliceParams.addParamsList.add(new IndexValues(offset + nextOldIndex, newValueBuffer));
      offset += newValueBuffer.length;
      newValueBuffer = new List();
    }

    void testRemove(int oldIndex) {
      if (oldIndex > nextOldIndex) {
        int count = oldIndex - nextOldIndex;
        spliceParams.removeParamsList.add(new IndexCount(nextOldIndex, count));
        offset -= count;
      }
    }

    for (int newIndex = 0; newIndex < newValues.length; ++newIndex) {
      V value = newValues[newIndex];
      int oldIndex = oldIndexMap[value];
      if ((oldIndex == null) || (oldIndex < nextOldIndex)) {
        newValueBuffer.add(value);
      } else {
        flush();
        testRemove(oldIndex);
        nextOldIndex = oldIndex + 1;
      }
    }
    flush();
    testRemove(oldValues.length);
    return spliceParams.isEmpty ? null : spliceParams;
  }

  List<IndexCount> detectFilter(List<V> oldValues, List<V> newValues) {
    List<IndexCount> removeParamsList = new List<IndexCount>();
    int oldIndex = 0;
    int oldLength = oldValues.length;
    int newLength = newValues.length;
    for (int newIndex = 0; newIndex <= newLength; ++newIndex) {
      V newValue = newValues[newIndex];
      int count = 0;
      while ((oldIndex + count < oldLength) && (oldValues[oldIndex + count] != newValue)) {
        ++count;
      }
      if (count != 0) {
        removeParamsList.add(new IndexCount(oldIndex, count));
      }
      oldIndex += count + 1;
    }
    return (removeParamsList.length != 0) ? removeParamsList : null;
  }

  List<int> detectReorder(List<V> oldValues, List<V> newValues) {
    List<int> indexList = new List();
    Map<V, int> newIndexMap = getIndexMap(newValues);
    for (int i = 0; i < oldValues.length; ++i) {
      indexList.add(newIndexMap[oldValues[i]]);
    }
    return isIdentity(indexList) ? null : indexList;
  }

  Map<V, int> getIndexMap(List<V> values) {
    Map<V, int> indexMap = new HashMap();
    for (int i = 0; i < values.length; ++i) {
      indexMap[values[i]] = i;
    }
    return indexMap;
  }

  bool isIdentity(List<int> list) {
    for (int i = 0; i < list.length; ++i) {
      if (list[i] != i) {
        return false;
      }
    }
    return true;
  }

  static List<int> invert(List<int> list) {
    List<int> result = new List.filled(list.length, 0);
    for (int i = 0; i < list.length; ++i) {
      result[list[i]] = i;
    }
    return result;
  }

  List<V> merge(List<List<V>> lists) {
    List<V> result = new List<V>();
    lists.forEach((List<V> list) { result.addAll(list); });
    return result;
  }

  int binarySearch(List<V> list, bool isBigEnough(V value)) {
    int length = list.length;
    int len2 = length >> 1;
    int step = 1;
    while (step <= len2) {
      step <<= 1;
    }
    int index = 0;
    while (step != 0) {
      if ((index + step <= length) && !isBigEnough(list[index + step - 1])) {
        index += step;
      }
      step >>= 1;
    }
    return index;
  }
}