part of insync;

class MapSpliceParams<K, V> {
  List<K> removedKeys;
  SimpleList<K> get $removedKeys => new SimpleList.adapter(removedKeys);

  Map<K, V> updatedValues;
  SimpleMap<K, V> get $updatedValues => new SimpleMap.adapter(updatedValues);

  MapSpliceParams() : removedKeys = new List<K>(), updatedValues = new HashMap<K, V>() {}

  bool get isEmpty => removedKeys.isEmpty && updatedValues.isEmpty;
}

class MapSpliceResult<K, V> {
  Map<K, V> removedValues;
  SimpleMap<K, V> get $removedValues => new SimpleMap.adapter(removedValues);

  Map<K, V> addedValues;
  SimpleMap<K, V> get $addedValues => new SimpleMap.adapter(addedValues);

  MapSpliceResult() : removedValues = new HashMap<K, V>(), addedValues = new HashMap<K, V>() {}
  MapSpliceResult.fromRemoved(this.removedValues) : addedValues = new HashMap<K, V>() {}

  bool get isEmpty => removedValues.isEmpty && addedValues.isEmpty;
}

/**
 * @todo Change this class to static when buddies will add generic method capability.
 */
class MapUtil<K, V> {
  HashMap<K, V> single(K key, V value) {
    HashMap<K, V> result = new HashMap<K, V>();
    result[key] = value;
    return result;
  }

  V putIfAbsent(Map<K, V> map, K key, V ifAbsent()) {
    V value = map[key];
    if ((value == null) && !map.containsKey(key)) {
      value = ifAbsent();
      map[key] = value;
    }
    return value;
  }

  bool any(Map<K, V> map, bool f(K key, V value)) {
    for (K key in map.keys) {
      if (!f(key, map[key])) {
        return false;
      }
    }
    return true;
  }

  Map<K, V> apply(Map<K, V> map, Map<K, V> other) {
    if (other == null) {
      return map;
    }
    for (K key in other.keys) {
      map[key] = other[key];
    }
    return map;
  }

  Wrap<V> trySet(Map<K, V> map, K key, V value) {
    assert(key != null);
    assert(value != null);
    V oldValue = map[key];
    if (oldValue == value) {
      return null;
    }
    map[key] = value;
    return new Wrap(oldValue);
  }

  MapSpliceResult<K, V> trySetAll(Map<K, V> map, Map<K, V> values) {
    assert(any(values, (key, value) => value != null));
    MapSpliceResult<K, V> spliceResult = new MapSpliceResult();
    for (K key in values.keys) {
      V value = values[key];
      Wrap<V> wrap = trySet(map, key, value);
      if (wrap == null) {
        continue;
      }
      V oldValue = wrap.value;
      if (oldValue != null) {
        spliceResult.removedValues[key] = oldValue;
      }
      spliceResult.addedValues[key] = value;
    }
    return !spliceResult.isEmpty ? spliceResult : null;
  }

  V tryRemove(Map<K, V> map, K key) {
    return map.remove(key);
  }

  Map<K, V> tryRemoveAll(Map<K, V> map, List<K> keys) {
    Map<K, V> items = new HashMap<K, V>();
    for (K key in keys) {
      V item = tryRemove(map, key);
      if (item != null) {
        items[key] = item;
      }
    }
    return items.isNotEmpty ? items : null;
  }

  Map<K, V> tryClear(Map<K, V> map) {
    if (map.isEmpty) {
      return null;
    }
    Map<K, V> items = new HashMap.from(map);
    map.clear();
    return items;
  }

  MapSpliceResult<K, V> trySplice(Map<K, V> map, List<K> removedKeys, Map<K, V> updatedValues) {
    assert(any(updatedValues, (key, value) => value != null));
    removedKeys = removedKeys.where((value) => map.containsKey(value)).toList(growable: false);
    Map<K, V> removedValues = tryRemoveAll(map, removedKeys);
    MapSpliceResult<K, V> spliceResult = trySetAll(map, updatedValues);
    if (spliceResult != null) {
      apply(spliceResult.removedValues, removedValues);
      return spliceResult;
    }
    return (removedValues != null) ? new MapSpliceResult.fromRemoved(removedValues) : null;
  }

  Map<K, K> tryReindex(Map<K, V> map, Map<K, K> keyMap) {
    assert(new MapUtil<K, K>().any(keyMap,
        (oldKey, newKey) => (newKey != null) && map.containsKey(oldKey)));
    Map<K, V> oldMap = tryClear(map);
    if (oldMap == null) {
      return null;
    }
    Map<K, K> resultMap = new HashMap<K, K>();
    for (K oldKey in oldMap.keys) {
      K newKey = keyMap[oldKey];
      if ((newKey == null) || (newKey == oldKey)) {
        assert(map[oldKey] == null);
        map[oldKey] = oldMap[oldKey];
      } else {
        assert(map[newKey] == null);
        map[newKey] = oldMap[oldKey];
        resultMap[oldKey] = newKey;
      }
    }
    return resultMap.isNotEmpty ? resultMap : null;
  }

  MapSpliceParams<K, V> detectSplice(Map<K, V> oldValues, Map<K, V> newValues) {
    assert(any(newValues, (key, value) => value != null));
    MapSpliceParams<K, V> params = new MapSpliceParams<K, V>();
    for (K key in oldValues.keys) {
      if (!newValues.containsKey(key)) {
        params.removedKeys.add(key);
      }
    }
    for (K key in newValues.keys) {
      V value = newValues[key];
      if (value != oldValues[key]) {
        params.updatedValues[key] = value;
      }
    }
    return !params.isEmpty ? params : null;
  }

  /**
   * New values must be unique by hashCode and equals.
   */
  Map<K, K> detectReindex(Map<K, V> oldValues, Map<K, V> newValues) {
    assert(any(newValues, (key, value) => value != null));
    Map<V, K> newValueKeys = new HashMap<V, K>();
    for (K key in newValues.keys) {
      newValueKeys[newValues[key]] = key;
    }
    Map<K, K> keyMap = new HashMap<K, K>();
    for (K oldKey in oldValues.keys) {
      K newKey = newValueKeys[oldValues[oldKey]];
      if (oldKey != newKey) {
        keyMap[oldKey] = newKey;
      }
    }
    return keyMap.isNotEmpty ? keyMap : null;
  }

  List<K> getRemovedKeys(Map<K, V> removedValues, Map<K, V> addedValues) {
    List<K> removedKeys = new List<K>();
    for (K key in removedValues.keys) {
      if (!addedValues.containsKey(key)) {
        removedKeys.add(key);
      }
    }
    return removedKeys;
  }

  bool equal(Map<K, V> x, Map<K, V> y) {
    if (identical(x, y)) {
      return true;
    }
    int length = y.length;
    for (K key in x.keys) {
      if ((--length < 0) || (x[key] != y[key])) {
        return false;
      }
    }
    return length == 0;
  }

  void setOrRemove(Map<K, V> map, K key, V value) {
    if (value != null) {
      map[key] = value;
    } else {
      map.remove(key);
    }
  }
}