part of insync;

class ObservableListMerger<V> extends AbstractListMerger<V> {
  ObservableListMerger(ObservableList<AbstractList<V>> source, {AbstractList<V> target}) :
    super(source, target: target) {
    own(source.onSplice.listen(_onSplice));
    own(source.onReplace.listen(_onReplace));
    own(source.onMove.listen(_onMove));
    own(source.onClear.listen(_onClear));
    own(source.onReorder.listen(_onReorder));
  }

  AbstractList<V> _createTarget() {
    return new ObservableList<V>();
  }

  List<int> _getIndexes(List<AbstractList<V>> lists) {
    int currentIndex = 0;
    int length = lists.length;
    List<int> indexes = new List<int>.filled(length + 1, null);
    for (int i = 0; i < length; ++i) {
      indexes[i] = currentIndex;
      currentIndex += lists[i].length;
    }
    indexes[length] = currentIndex;
    return indexes;
  }

  // 0 x,x
  // 2 x,x,x delete
  // 5 x,x,x,x
  // 9 x,x

  // 0 x,x
  // 2 x,x,x,x
  // 6 x,x

  void _onSplice(ObservableListSpliceEventParams<AbstractList<V>> params) {
    ListSpliceResult<AbstractList<V>> spliceResult = params.spliceResult;
    List<int> indexes = _getIndexes(spliceResult.oldValues);
    List<IndexCount> removeParamsList = spliceResult.removedValuesList.map((IndexValues<AbstractList<V>> indexValues) {
      return new IndexCount(indexes[indexValues.index], _count(indexValues.values));
    }).toList(growable: false);
    spliceResult.removedValuesList.forEach((IndexValues<AbstractList<V>> indexValues) {
      indexes.removeRange(indexValues.index, indexValues.index + indexValues.values.length);
      int count = _count(indexValues.values);
      for (int i = indexValues.index; i < indexes.length; ++i) {
        indexes[i] -= count;
      }
    });
    List<IndexValues<AbstractList<V>>> addParamsList = spliceResult.addedValuesList.map((IndexValues<AbstractList<V>> indexValues) {
      return new IndexValues<AbstractList<V>>(indexes[indexValues.index], _merge(indexValues.values));
    }).toList(growable: false);
    _target.trySplice(removeParamsList, addParamsList);
  }

  void _onReplace(ObservableListReplaceEventParams<AbstractList<V>> params) {
    int index = _count(_source.list, 0, params.index);
    _target.trySplice(
        [new IndexCount(index, params.oldValue.length)],
        [new IndexValues<V>(index, params.newValue.list)]);
  }

  void _onMove(ObservableListMoveEventParams<AbstractList<V>> params) {
    int count = params.value.length;
    List<int> indexes = new List<int>.filled(_target.length, null);
    int currentIndex = 0;

    void shiftList(int listLength, int shift) {
      for (int j = 0; j < listLength; ++j) {
        indexes[currentIndex] = currentIndex + shift;
        ++currentIndex;
      }
    }

    for (int i = 0, l = min(params.fromIndex, params.toIndex); i < l; ++i) {
      shiftList(_source[i].length, 0);
    }
    if (params.fromIndex <= params.toIndex) {
      // [1], [2], [3], [4], [5]        [2] move to 3
      // [1], [3], [4], [2], [5]
      shiftList(count, _count(_source.list, params.fromIndex, params.toIndex - params.fromIndex));
      for (int i = params.fromIndex; i < params.toIndex; ++i) {
        shiftList(_source[i].length, -count);
      }
    } else {
      // [1], [2], [3], [4], [5]        [4] move to 1
      // [1], [4], [2], [3], [5]
      for (int i = params.toIndex + 1; i <= params.fromIndex; ++i) {
        shiftList(_source[i].length, count);
      }
      shiftList(count, -_count(_source.list, params.toIndex + 1, params.fromIndex - params.toIndex));
    }
    for (int i = max(params.fromIndex, params.toIndex) + 1, l = _source.length; i < l; ++i) {
      shiftList(_source[i].length, 0);
    }
    _target.tryReorder(indexes);
  }

  void _onClear(List<AbstractList<V>> values) {
    _target.tryClear();
  }

  void _onReorder(ObservableListReorderEventParams<AbstractList<V>> params) {
    List<int> oldIndexes = _getIndexes(params.values);
    List<int> newIndexes = _getIndexes(_source.list);
    List<int> indexes = new List<int>.filled(_target.length, null);
    for (int i = 0, l = params.values.length; i < l; ++i) {
      AbstractList<V> list = params.values[i];
      int oldIndex = oldIndexes[i];
      int newIndex = newIndexes[params.indexList[i]];
      for (int j = 0, m = list.length; j < m; ++j) {
        indexes[oldIndex + j] = newIndex + j;
      }
    }
    _target.tryReorder(indexes);
  }
}