part of insync;

class _ObservableListMergerBunch<V> extends _AbstractListMergerBunch<V> {
  AbstractList<AbstractList<V>> _source;
  AbstractList<V> _target;
  ObservableList<V> _list;

  _ObservableListMergerBunch(ObservableListMerger<V> merger, this._list) :
    _source = merger.source, _target = merger.target {
    own(_list.onSplice.listen(_onSplice));
    own(_list.onReplace.listen(_onReplace));
    own(_list.onMove.listen(_onMove));
    own(_list.onClear.listen(_onClear));
    own(_list.onReorder.listen(_onReorder));
  }

  int _getIndex() {
    List<AbstractList<V>> lists = _source.list;
    int index = 0;
    for (int i = 0, l = lists.length; i < l; ++i) {
      AbstractList<V> list = lists[i];
      if (identical(list, _list)) {
        return index;
      }
      index += list.length;
    }
    throw new Exception("ObservableListMerger object is corrupted. "
        "Please avoid external modification of target list.");
  }

  void _onSplice(ObservableListSpliceEventParams<V> params) {
    ListSpliceResult<V> spliceResult = params.spliceResult;
    int index = _getIndex();
    List<IndexCount> removeParamsList = spliceResult.removedValuesList.map((IndexValues<V> indexValues) {
      return new IndexCount(indexValues.index + index, indexValues.values.length);
    }).toList(growable: false);
    List<IndexValues<V>> addParamsList = spliceResult.addedValuesList.map((IndexValues<V> indexValues) {
      return new IndexValues<V>(indexValues.index + index, indexValues.values.toList(growable: false));
    }).toList(growable: false);
    _target.trySplice(removeParamsList, addParamsList);
  }

  void _onReplace(ObservableListReplaceEventParams<V> params) {
    _target.trySet(params.newValue, _getIndex() + params.index);
  }

  void _onMove(ObservableListMoveEventParams<V> params) {
    int index = _getIndex();
    _target.tryMove(index + params.fromIndex, index + params.toIndex);
  }

  void _onClear(List<V> values) {
    _target.tryRemoveAll(_getIndex(), values.length);
  }

  void _onReorder(ObservableListReorderEventParams<V> params) {
    int index = _getIndex();
    List<int> listIndexList = params.indexList;
    int listLength = listIndexList.length;
    int targetLength = _target.length;
    List<int> targetIndexList = new List<int>.filled(targetLength, null);
    for (int i = 0; i < index; ++i) {
      targetIndexList[i] = i;
    }
    for (int i = 0; i < listLength; ++i) {
      targetIndexList[index + i] = index + listIndexList[i];
    }
    for (int i = index + listLength; i < targetLength; ++i) {
      targetIndexList[i] = i;
    }
    _target.tryReorder(targetIndexList);
  }
}