part of insync;

class ObservableListReverser<V> extends AbstractListReverser<V> {
  ObservableListReverser(ObservableList<V> source, {AbstractList<V> target}) :
    super(source, target: target) {
    own(source.onSplice.listen(_onSplice));
    own(source.onReplace.listen(_onReplace));
    own(source.onMove.listen(_onMove));
    own(source.onClear.listen(_onClear));
    own(source.onReorder.listen(_onReorder));
  }

  void _onSplice(ObservableListSpliceEventParams<V> params) {
    ListSpliceResult<V> spliceResult = params.spliceResult;
    int oldLength = _target.length;
    int newLength = oldLength;

    List<IndexCount> removeParamsList = spliceResult.removedValuesList.map((IndexValues<V> indexValues) {
      int length = indexValues.values.length;
      int index = oldLength - indexValues.index - length;
      newLength -= length;
      return new IndexCount(index, length);
    }).toList(growable: false).reversed.toList(growable: false);

    List<IndexValues<V>> addedValuesList =
        spliceResult.addedValuesList.reversed.toList(growable: false);
    addedValuesList.forEach((IndexValues<V> indexValues) {
      newLength += indexValues.values.length;
    });

    List<IndexValues<V>> addParamsList = addedValuesList.map((IndexValues<V> indexValues) {
      List<V> values = indexValues.values;
      int length = values.length;
      int index = newLength - indexValues.index - length;
      return new IndexValues<V>(index, values.reversed.toList(growable: false));
    }).toList(growable: false);

    _target.trySplice(removeParamsList, addParamsList);
  }

  void _onReplace(ObservableListReplaceEventParams<V> params) {
    _target.trySet(params.newValue, _target.length - params.index - 1);
  }

  void _onMove(ObservableListMoveEventParams<V> params) {
    _target.tryMove(
        _target.length - params.fromIndex - 1,
        _target.length - params.toIndex - 1);
  }

  void _onClear(List<V> values) {
    _target.tryClear();
  }

  void _onReorder(ObservableListReorderEventParams<V> params) {
    List<int> indexList = params.indexList;
    int length = indexList.length;
    List<int> indexes = new List<int>.filled(length, null);
    for (int i = 0; i < length; ++i) {
      indexes[length - i - 1] = length - indexList[i] - 1;
    }
    _target.tryReorder(indexes);
  }
}