part of insync;

class ObservableSet<V> extends AbstractSet<V> implements ObservableCollection<V> {
  EventDispatcher<SetSpliceResult<V>> onSplice = new EventDispatcher();
  EventDispatcher<List<V>> onClear = new EventDispatcher();
  EventDispatcher onChange = new EventDispatcher();

  Property<int> _length;
  Property<int> get lengthProperty => _length;

  ObservableSet() : super(), _length = new Property<int>(0);

  ObservableSet.from(List<V> values) : super.from(values),
  _length = new Property<int>(values.length);

  ObservableSet.adapter(Set<V> set) : super.adapter(set), _length = new Property<int>(set.length);

  void destroy() {
    onSplice.destroy();
    onClear.destroy();
    onChange.destroy();
    _length.destroy();
    super.destroy();
  }

  List<V> tryClear() {
    List<V> oldValues = super.tryClear();
    if (oldValues == null) {
      return null;
    }
    _length.value = 0;
    onClear.fire(oldValues);
    onChange.fire();
    return oldValues;
  }

  SetSpliceResult<V> trySplice(List<V> removedValues, List<V> addedValues) {
    SetSpliceResult<V> spliceResult = super.trySplice(removedValues, addedValues);
    if (spliceResult == null) {
      return null;
    }
    _length.value = length;
    onSplice.fire(spliceResult);
    onChange.fire();
    return spliceResult;
  }

  ObservableSet<dynamic> createEmpty() {
    return new ObservableSet<dynamic>();
  }

  ObservableList<dynamic> createEmptyList() {
    return new ObservableList<dynamic>();
  }

  ObservableMap<dynamic, dynamic> createEmptyMap() {
    return new ObservableMap<dynamic, dynamic>();
  }

  ObservableSet<dynamic> createEmptySet() {
    return new ObservableSet<dynamic>();
  }

  ObservableSetMapper<V, dynamic> createMapper({dynamic create(V data),
    void destroy(dynamic item, V data)}) {
    return new ObservableSetMapper<V, dynamic>(this, create: create, destroy: destroy);
  }

  ObservableSetMapper<V, dynamic> createMapperTo({AbstractSet<dynamic> target,
    dynamic create(V data), void destroy(dynamic item, V data)}) {
    return new ObservableSetMapper<V, dynamic>(this, target: target, create: create,
        destroy: destroy);
  }

  ObservableSetFilterer<V> createFilterer({bool filter(V value)}) {
    return new ObservableSetFilterer<V>(this, filter: filter);
  }

  ObservableSetFilterer<V> createFiltererTo({AbstractSet<V> target, bool filter(V value)}) {
    return new ObservableSetFilterer<V>(this, target: target, filter: filter);
  }

  ObservableSetObserver<V> createObserver({void add(V value), void remove(V value),
    void clear(List<V> values), void change()}) {
    return new ObservableSetObserver<V>(this, add: add, remove: remove, clear: clear,
        change: change);
  }

  ObservableSetOrderer<V> createOrderer({AbstractList<V> target}) {
    return new ObservableSetOrderer<V>(this, target: target);
  }

  ObservableSetSorterComparing<V> createSorterComparing({AbstractList<V> target,
    int compare(V x, V y)}) {
    return new ObservableSetSorterComparing<V>(this, target: target, compare: compare);
  }

  ObservableSetIndexer<dynamic, V> createIndexer(
      {AbstractMap<dynamic, V> target, dynamic getKey(V value)}) {
    return new ObservableSetIndexer<dynamic, V>(this, target: target, getKey: getKey);
  }

  ObservableSetHeaper<V> createHeaper({AbstractSet<V> target}) {
    return new ObservableSetHeaper<V>(this, target: target);
  }
}