part of insync;

class PropertyMapper<T> extends Owner {
  final List<Property> sources;
  final Function _createValue;
  final Function _destroyValue;
  final bool _acceptNull;
  List _sourceValues;
  T _targetValue;

  Property<T> _target;
  Property<T> get target => _target;

  PropertyMapper(this.sources,
      {Function create, Function destroy, Property<T> target, bool acceptNull}) :
      _createValue = create, _destroyValue = destroy, _acceptNull = acceptNull {
    _target = (target != null) ? target : new Property<T>()..ownBy(this);
    _update();
    sources.forEach(watch);
  }

  void destroy() {
    T oldValue = _target.value;
    if (oldValue == _targetValue) {
      _target.value = null;
    }
    _done();
    _sourceValues = null;
    super.destroy();
  }

  void listen(EventDispatcher dispatcher) {
    own(dispatcher.listen(_update));
  }

  void watch(Property property) {
    listen(property.onChange);
  }

  void _update([dynamic p]) {
    List sourceValues = sources.map((Property source) => source.value).toList();
    T targetValue;
    if (_acceptNull || sourceValues.every((value) => value != null)) {
      targetValue = Function.apply(_createValue, sourceValues);
    } else {
      targetValue = null;
      sourceValues = null;
    }
    target.value = targetValue;
    _done();
    _targetValue = targetValue;
    _sourceValues = sourceValues;
  }

  void _done() {
    if ((_destroyValue != null) && (_sourceValues != null)) {
      List args = [_targetValue];
      args.addAll(_sourceValues);
      Function.apply(_destroyValue, args);
    }
  }
}