part of insync;

class Component extends Owner {
  Component _parent;
  Component get parent => _parent;

  bool _wasAfterAppend = false;
  bool get wasAfterAppend => _wasAfterAppend;

  Element _el;
  Element get el => _el;

  Element _replacedEl;
  Element get replacedEl => _replacedEl;

  ObservableMap<String, Component> _children;
  ObservableMap<String, Component> get children => _children;

  Map<String, Element> _rawEls;
  ComponentElements _els;
  ComponentElements get elements => _els;

  ObservableMapMapper<String, Component, _ComponentChild> _childMapper;
  ObservableMapInserter<String, _ComponentChild> _childInserter;
  List<ComponentReplaceable> _replaceables;
  List<ComponentList> _lists;

  String get templateId => 'Component';
  String get template => '<div></div>';
  Map<String, Function> get renderers => {'root': renderRoot};

  void destroy() {
    if (_parent != null) {
      throw new Exception('insync.Component.destroy must be used for root or removed components only');
    }
    if (_wasAfterAppend) {
      releaseDom();
    }
    if (_el != null) {
      _el.remove();
      List<ComponentList> lists = _lists.toList(growable: false);
      lists.forEach(Destroyable.destroyObject);
      _lists = null;
      List<ComponentReplaceable> replaceables = _replaceables.toList(growable: false);
      replaceables.forEach(Destroyable.destroyObject);
      _replaceables = null;

      _childInserter.destroy();
      _childInserter = null;
      _childMapper.destroy();
      _childMapper = null;

      unrender();

      _children.destroy();
      _children = null;
    }
    _els = null;
    _rawEls = null;
    _el = null;
    afterDestroy();
    super.destroy();
  }

  void beforeRender() {}
  dynamic renderRoot(Element el) {}
  void afterRender() {}
  void afterAppend() {}
  void releaseDom() {}
  void unrender() {}
  void afterDestroy() {}
  void layout() {}

  void render([Element replacedEl]) {
    if (_el != null) {
      return;
    }
    _replacedEl = replacedEl;
    TemplateOutput output = TemplateRegistry.createElement(templateId, template);
    _el = output.root;
    _rawEls = output.els;
    _els = new ComponentElements(_rawEls);
    _children = new ObservableMap();
    _replaceables = new List<ComponentReplaceable>();
    _lists = new List<ComponentList>();
    _childMapper = _children.createMapper(
        create: (Component component) => new _ComponentChild(this, component),
        destroy: Destroyable.destroyFromMapper
    );
    _childInserter = _childMapper.target.createInserter(
        add: (_ComponentChild child, String key) { child.attach(key); },
        remove: (_ComponentChild child, String key) { child.detach(); }
    );
    beforeRender();
    Map<String, Element> els = new Map.from(_rawEls);
    Map<String, Function> renderers = this.renderers;
    els.forEach((String id, Element el) {
      Function renderer = renderers[id];
      if (renderer == null) {
        return;
      }
      var result = renderer(els[id]);
      if (identical(el, _el)) {
        if (result is AbstractList) {
          addList(result, _el);
        }
      } else {
        if (result is Component) {
          _children[id] = result;
        } else if (result is Property) {
          addReplaceable(result, id);
        } else if (result is AbstractList) {
          addList(result, _el);
        } else if ((result is bool) && !result) {
          _els.remove(id);
        }
      }
    });
    afterRender();
  }

  void renderTo(Element el) {
    render();
    el.children.add(_el);
    _afterAppend();
  }

  /**
   * @todo Test event listeners persistance
   */
  void renderAs(Element el) {
    render(el);
    el.replaceWith(_el);
    _el.classes.addAll(el.classes);
    _afterAppend();
  }

  void invokeLayout() {
    if (wasAfterAppend) {
      layout();
    }
  }

  /**
   * @todo Test event listeners persistance
   */
  void remove() {
    if (_parent != null) {
      throw new Exception('insync.Component.remove must be used for root components only');
    }
    _el.remove();
  }

  ComponentReplaceable addReplaceable(Property<Component> source, String id) {
    return new ComponentReplaceable(this, source, id);
  }

  ComponentList addList(AbstractList<Component> source, [Element el]) {
    if (el == null) {
      el = _el;
    }
    return new ComponentList(this, source, el);
  }

  void _afterAppend() {
    if (_wasAfterAppend || (_el == null)) {
      return;
    }
    if ((_parent != null) && !_parent._wasAfterAppend) {
      return;
    }
    if ((_parent == null) && (ElementUtil.queryParent(_el, 'body') == null)) {
      return;
    }
    _wasAfterAppend = true;
    afterAppend();
    _children.each(_childAfterAppend);
    _lists.forEach((ComponentList list) => list.source.each(_childAfterAppend));
  }

  void _initChild(Component component, [Element replacedEl]) {
    component.render(replacedEl);
    component._parent = this;
  }

  void _doneChild(Component component) {
    component._parent = null;
  }

  static void _childAfterAppend(Component component) => component._afterAppend();
}